# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import typing
import weakref
from typing import Optional, Type

import click
from click_option_group import GroupedOption, MutuallyExclusiveOptionGroup, OptionGroup, RequiredAllOptionGroup, \
    optgroup


class _OptionGroupMixin:

    def given_option_with_count(self, ctx: click.Context, opts: dict) -> typing.Tuple[typing.Set[str], int, int]:
        option_names = set(self.get_option_names(ctx))
        given_option_names = option_names.intersection(opts)
        return given_option_names, len(given_option_names), len(option_names)

    def option_count_requirement_met(self, ctx: click.Context, opts: dict) -> bool:
        for grp in self._extgrp._groups:
            if not grp._option_group().option_count_requirement_met(ctx, opts):
                return False

        return True


class _OptionGroup(OptionGroup, _OptionGroupMixin):
    def __init__(self, name: typing.Optional[str] = None, *,
                 hidden=False, help: Optional[str] = None, extgrp: '_ExtGroup' = None) -> None:
        super().__init__(name, hidden=hidden, help=help)
        self._extgrp = extgrp
        extgrp._set_option_group(self)


class _RequiredAllOptionGroup(RequiredAllOptionGroup, _OptionGroupMixin):
    def __init__(self, name: typing.Optional[str] = None, *,
                 hidden=False, help: Optional[str] = None, extgrp: '_ExtGroup' = None) -> None:
        super().__init__(name, hidden=hidden, help=help)
        self._extgrp = extgrp
        extgrp._set_option_group(self)

    def option_count_requirement_met(self, ctx: click.Context, opts: dict):
        option_names = set(self.get_options(ctx))

        if not _OptionGroupMixin.option_count_requirement_met(self, ctx, opts):
            return False

        return option_names.issubset(opts)

    def handle_parse_result(self, option: GroupedOption, ctx: click.Context, opts: dict) -> None:
        if self._extgrp._option_count_requirement_met(ctx, opts):
            return

        option_names = set(self.get_options(ctx))

        if not option_names.issubset(opts):
            group_name = self._group_name_str()
            required_names = option_names.difference(option_names.intersection(opts))
            option_info = self.get_error_hint(ctx, required_names)

            raise click.UsageError(
                f"Missing required options from {group_name} option group:\n{option_info}",
                ctx=ctx
            )


class _MutuallyExclusiveOptionGroup(MutuallyExclusiveOptionGroup, _OptionGroupMixin):
    def __init__(self, name: typing.Optional[str] = None, *,
                 hidden=False, help: Optional[str] = None, extgrp: '_ExtGroup' = None) -> None:
        super().__init__(name, hidden=hidden, help=help)
        self._extgrp = extgrp
        extgrp._set_option_group(self)

    def option_count_requirement_met(self, ctx: click.Context, opts: dict) -> bool:
        option_names = set(self.get_options(ctx))
        given_option_names = option_names.intersection(opts)
        if len(given_option_names) > 1:
            return False

        if len(given_option_names) == 0:
            count = 0
            for grp in self._extgrp._groups:
                count += int(grp._option_group().option_count_requirement_met(ctx, opts))
            if count > 1:
                return False

        return True

    def handle_parse_result(self, option: GroupedOption, ctx: click.Context, opts: dict) -> None:
        if self._extgrp._option_count_requirement_met(ctx, opts):
            return

        option_names = set(self.get_options(ctx))
        given_option_names = option_names.intersection(opts)
        given_option_count = len(given_option_names)

        for grp in self._extgrp._groups:
            names, given, _ = grp._option_group().given_option_with_count(ctx, opts)
            given_option_count += given
            given_option_names = given_option_names.union(names)

        if given_option_count > 1:
            group_name = self._group_name_str()
            option_info = self.get_error_hint(ctx, given_option_names)

            raise click.UsageError(
                f"Mutually exclusive options from {group_name} option group "
                f"cannot be used at the same time:\n{option_info}",
                ctx=ctx
            )


class _RequiredMutuallyExclusiveOptionGroup(MutuallyExclusiveOptionGroup, _OptionGroupMixin):
    def __init__(self, name: typing.Optional[str] = None, *,
                 hidden=False, help: Optional[str] = None, extgrp: '_ExtGroup' = None) -> None:
        super().__init__(name, hidden=hidden, help=help)
        self._extgrp: _ExtGroup = extgrp
        extgrp._set_option_group(self)

    @property
    def name_extra(self) -> typing.List[str]:
        return super().name_extra + ['required']

    def option_count_requirement_met(self, ctx: click.Context, opts: dict) -> bool:
        option_names = set(self.get_options(ctx))
        given_option_names = option_names.intersection(opts)
        if len(given_option_names) > 1:
            return False

        if len(given_option_names) == 0:
            count = 0
            for grp in self._extgrp._groups:
                count += int(grp._option_group().option_count_requirement_met(ctx, opts))
            if count != 1:
                return False

        return True

    def handle_parse_result(self, option: GroupedOption, ctx: click.Context, opts: dict) -> None:
        if self._extgrp._option_count_requirement_met(ctx, opts):
            return

        super().handle_parse_result(option, ctx, opts)

        option_names = set(self.get_option_names(ctx))
        given_option_names = option_names.intersection(opts)

        for grp in self._extgrp._groups:
            names, given, _ = grp._option_group().given_option_with_count(ctx, opts)
            given_option_names = given_option_names.union(names)

        if len(given_option_names) == 0:
            group_name = self._group_name_str()
            option_info = self.get_error_hint(ctx)
            if self._extgrp._groups:
                option_info += '\n  ' + 'one or more groups, check --help'

            raise click.UsageError(
                "Missing one of the required mutually exclusive options from "
                f"{group_name} option group:\n{option_info}",
                ctx=ctx
            )


class _Base:
    def __init__(self):
        self._callbacks: typing.List[callable] = []
        self._last_callbacks: typing.List[callable] = []

    def _add_option(self, decorator: callable, *args, **kwargs):
        if 'dest' in kwargs:
            args = (*args, kwargs['dest'])
            del kwargs['dest']
        if 'choices' in kwargs:
            kwargs['type'] = click.Choice(kwargs['choices'])
            del kwargs['choices']

        self._callbacks.append(lambda: decorator(*args, **kwargs))

    def add_args_to_func(self, f: callable):
        for c in reversed(self._last_callbacks):
            f = c()(f)
        for c in reversed(self._callbacks):
            f = c()(f)

        return f

    def add_custom_decorator(self, decorator: callable, *dec_args, **dec_kwargs):
        self._callbacks.append(lambda: decorator(*dec_args, **dec_kwargs))


class _Group(_Base):
    def __init__(self, name: Optional[str] = None, *,
                 help: Optional[str] = None,
                 cls: Optional[Type[OptionGroup]] = None,
                 **kwargs):
        super().__init__()
        self._cls = cls
        self._callbacks.append(lambda: optgroup.group(name, help=help, cls=cls, **kwargs))

    def add_option(self, *args, **kwargs):
        """@see click_option_group.optgroup.option"""
        self._add_option(optgroup.option, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.add_args_to_func


class _ExtGroup(_Group):
    def __init__(self, name: Optional[str] = None, *,
                 help: Optional[str] = None,
                 cls: Optional[Type[OptionGroup]] = None,
                 parent: typing.Optional['_ExtGroup'] = None):
        super().__init__(name, help=help, cls=cls, extgrp=self)
        self._parent = weakref.ref(parent) if parent else None
        self._groups = []
        self._option_group = None

    def _set_option_group(self, o: OptionGroup):
        self._option_group = weakref.ref(o)

    def _option_count_requirement_met(self, ctx: click.Context, opts: dict) -> bool:
        if self._parent:
            return self._parent()._option_count_requirement_met(ctx, opts)
        else:
            return self._option_group().option_count_requirement_met(ctx, opts)


class _ExtOptionGroup(_ExtGroup):
    def __init__(self, name: Optional[str] = None, *,
                 help: Optional[str] = None,
                 cls: Optional[Type[OptionGroup]] = None,
                 parent: typing.Optional['_ExtGroup'] = None):
        super().__init__(name, help=help, cls=cls, parent=parent)

    def add_mutually_exclusive_group(self, name: Optional[str] = None, *,
                                     help: Optional[str] = None,
                                     required: bool = False
                                     ) -> '_ExtMutuallyExclusiveGroup':
        if self._cls == _RequiredAllOptionGroup:
            required = True
        g = _ExtMutuallyExclusiveGroup(name, help=help,
                                       cls=_RequiredMutuallyExclusiveOptionGroup if required else _MutuallyExclusiveOptionGroup,
                                       parent=self)
        self._last_callbacks.append(g)
        self._groups.append(g)
        return g


class _ExtMutuallyExclusiveGroup(_ExtGroup):
    def __init__(self, name: Optional[str] = None, *,
                 help: Optional[str] = None,
                 cls: Optional[Type[OptionGroup]] = None,
                 parent: typing.Optional['_ExtGroup'] = None):
        super().__init__(name, help=help, cls=cls, parent=parent)

    def add_group(self, name: Optional[str] = None, *,
                  help: Optional[str] = None, require_all=False) -> '_ExtOptionGroup':
        g = _ExtOptionGroup(name, help=help, cls=_RequiredAllOptionGroup if require_all else _OptionGroup, parent=self)
        self._last_callbacks.append(g)
        self._groups.append(g)
        return g


class OptionContext(_Base):

    def add_option(self, *args, **kwargs):
        """
        @see click.option decorator for details of parameters
        """
        self._add_option(click.option, *args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """
        @see click.argument decorator for details of parameters.

        The help parameter is ignored, it's not supported by click v8,
        but it can be passed to this method.
        """
        if 'help' in kwargs:
            del kwargs['help']
        if 'choices' in kwargs:
            kwargs['type'] = click.Choice(kwargs['choices'])
            del kwargs['choices']

        self._callbacks.append(lambda: click.argument(*args, **kwargs))

    def add_group(self, name: Optional[str] = None, *,
                  help: Optional[str] = None,
                  cls: Optional[Type[OptionGroup]] = None) -> _Group:
        g = _Group(name, help=help, cls=cls)
        self._callbacks.append(g)
        return g

    def add_multilevel_group(self, name: Optional[str] = None, *,
                             help: Optional[str] = None, require_all=False) -> _ExtOptionGroup:
        g = _ExtOptionGroup(name, help=help, cls=_RequiredAllOptionGroup if require_all else _OptionGroup)
        self._callbacks.append(g)
        return g

    def add_mutually_exclusive_group(self, name: Optional[str] = None, *,
                                     help: Optional[str] = None,
                                     required: bool = False
                                     ) -> _ExtMutuallyExclusiveGroup:
        g = _ExtMutuallyExclusiveGroup(name, help=help,
                                       cls=_RequiredMutuallyExclusiveOptionGroup if required else _MutuallyExclusiveOptionGroup)
        self._callbacks.append(g)
        return g

    def register_args(self, arg_reg_func: callable):
        arg_reg_func(self)

    def add_args_to_func(self, f: callable):
        for c in reversed(self._callbacks):
            f = c()(f)

        return f



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import os

from dewi_core.application import Application
from . import ReviewCommand


def main():
    if int(os.environ.get('DEWI_CORE_DEV_WITH_PLUGINS', '0')) == 1:
        app = Application('dewi-review')
        app.load_plugin('dewi_core.commands.review.ReviewPlugin')
    else:
        app = Application('dewi-review', ReviewCommand)
    app.run()


if __name__ == '__main__':
    main()



import collections

from dewi_core.logger import log_debug


def log_debug_dict(c: dict, level=0):
    for k, v in c.items():
        if isinstance(v, collections.abc.Mapping):
            log_debug(f'{" " * level}{k}:')
            log_debug_dict(v, level + 2)
        else:
            log_debug(f'{" " * level}{k}: {v}')



#  Copyright 2015-2022, Laszlo Attila Toth
#  Distributed under the terms of the Apache License, Version 2.0

import os
import sys
import time
import typing

import click

from dewi_core.appcontext import ApplicationContext
from dewi_core.command import Command
from dewi_core.commandregistry import CommandRegistry
from dewi_core.config.node import Node
from dewi_core.config_env import ConfigDirRegistry, EnvConfig
from dewi_core.loader.loader import PluginLoader
from dewi_core.logger import LogLevel, LoggerConfig, create_logger_from_config, log_debug
from dewi_core.optioncontext import OptionContext
from dewi_core.utils.exception import print_backtrace
from dewi_core.utils.levenshtein import get_similar_names_to
from dewi_core.utils.time import humanize_time

try:
    from click._unicodefun import _verify_python_env

    def _dummy():
        pass


    # the UTF-8 check is buggy, ignore it
    _verify_python_env.__code__ = _dummy.__code__
except ImportError:
    pass

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class AliasedGroup(click.Group):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._command_registry: CommandRegistry = CommandRegistry()

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._command_registry:
            ctx.invoked_subcommand_ = cmd_name
            return super().get_command(ctx,
                                       self._command_registry.get_command_class_descriptor(cmd_name).get_class().name)
        elif not self._command_registry.get_command_count():
            rv = super().get_command(ctx, cmd_name)
            if rv is not None:
                return rv
            else:
                ctx.fail(f'Unknown command name {cmd_name}')

        print(f"ERROR: The command '{cmd_name}' is not known.\n")
        similar_names = get_similar_names_to(cmd_name, sorted(self._command_registry.get_command_names()))

        print('Similar names - firstly based on command name length:')
        for name in similar_names:
            print('  {:30s}   -- {}'.format(
                name,
                self._command_registry.get_command_class_descriptor(name).get_class().description))

        return click.core.Command('dummy', callback=lambda: sys.exit(1))

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args

    def add_command(self, cmd, name=None):
        if hasattr(cmd.callback, 'command_class'):
            self._command_registry.register_class(cmd.callback.command_class)

        super().add_command(cmd, name)


def _wait_for_termination_if_needed(ctx: ApplicationContext):
    if 'wait' in ctx.args and ctx.args.wait:
        print("\nPress ENTER to continue")
        input("")


def _print_exception(ctx: ApplicationContext, exc: BaseException):
    print_bt = 'print_backtraces' in ctx.args and ctx.args.print_backtraces
    if print_bt or os.environ.get('DEWI_DEBUG', 0) == '1':
        print_backtrace()
    print(f'Exception: {exc} (type: {type(exc).__name__})', file=sys.stderr)


def _app_main(parent_cls):
    def app_main(self, *args, **kwargs):
        start = end = 0

        def print_run_time():
            nonlocal end
            end = time.time()
            if os.environ.get('DEWI_SHOW_RUN_TIME', '0') == '1':
                diff = int((end - start) * 100) / 100
                print(f'Runtime: {humanize_time(end - start, format=True)} ({diff:2} seconds)')

        start = time.time()
        try:
            parent_cls.main(self, *args, **kwargs)
        except SystemExit:
            print_run_time()
            _wait_for_termination_if_needed(self.callback.app_context)
            raise
        except BaseException as exc:
            print_run_time()
            _print_exception(self.callback.app_context, exc)
        _wait_for_termination_if_needed(self.callback.app_context)
        sys.exit(1)

    return app_main


class AppGroup(AliasedGroup):
    main = _app_main(AliasedGroup)


class AppCommand(click.Command):
    main = _app_main(click.Command)


def get_invoked_subommand(ctx: click.Context):
    if hasattr(ctx, 'invoked_subcommand_'):
        return ctx.invoked_subcommand_
    else:
        return ctx.invoked_subcommand


def get_command_from_plugin_appcontext(plugin_name: str, ctx: ApplicationContext) -> typing.Type[Command]:
    """
    Returns the command  class specified in the ctx ApplicationContext via its running_command_
    and running_subcommands_ members.

    The ns comes from an already parsed namespace, especially via
    dewi_core.remoting.serialize_argparse_namespace and deserialize_argparse_namespace methods.

    The command must exist.

    :param plugin_name: The command plugin or app plugin containing the command referrered in ctx
    :param ctx: the prepared ApplicationContext object
    :return: the command class based on ctx.
    """
    command_registry = CommandRegistry()
    cfg_dir_registry = ConfigDirRegistry(EnvConfig(ctx.environment))
    cfg_dir_registry.register_config_directories(ctx.config_directories)
    loader = PluginLoader(command_registry, cfg_dir_registry)
    loader.load([plugin_name])
    cfg_dir_registry.load_env()
    cmd_class = command_registry.get_command_class_descriptor(ctx.command_names.command).get_class()

    if ctx.command_names.subcommands:
        for sub_cmd_name in ctx.command_names.subcommands:
            for cc in cmd_class.subcommand_classes:
                if cc.name == sub_cmd_name:
                    cmd_class = cc

    return cmd_class


def run_command_from_appcontext(plugin_name: str, ctx: ApplicationContext) -> typing.Optional[int]:
    """
    Runs the command returned by get_command_from_plugin_appcontext(). See that function.
    """
    return get_command_from_plugin_appcontext(plugin_name, ctx)().run(ctx)


def _list_commands(prog_name: str, command_registry: CommandRegistry, *, all_commands: bool = False):
    commands = dict()
    max_length = 0
    infix = '  - alias of '

    for name in command_registry.get_command_names():
        command_name, description = _get_command_name_and_description(command_registry, name)

        if name == command_name:
            cmdname = name
        else:
            if not all_commands:
                continue

            cmdname = (name, command_name)

        if len(name) > max_length:
            max_length = len(name)

        commands[name] = (cmdname, description)

    if all_commands:
        format_str = "  {0:<" + str(max_length * 2 + len(infix)) + "}   -- {1}"
    else:
        format_str = "  {0:<" + str(max_length) + "}   -- {1}"

    alias_format_str = "{0:<" + str(max_length) + "}" + infix + "{1}"

    print(f'Available {prog_name.capitalize()} Commands.')
    for name in sorted(commands):
        cmdname, description = commands[name]
        if isinstance(cmdname, tuple):
            cmdname = alias_format_str.format(*cmdname)
        print(format_str.format(cmdname, description))


def _get_command_name_and_description(command_registry, name):
    desc = command_registry.get_command_class_descriptor(name)
    description = desc.get_class().description
    command_name = desc.get_name()
    return command_name, description


class _ListAllCommand(Command):
    name = 'list-all'
    description = 'Lists all available command with aliases'

    def run(self, ctx: ApplicationContext):
        _list_commands(ctx.program_name, ctx.command_registry, all_commands=True)


class _ListCommand(Command):
    name = 'list'
    description = 'Lists all available command with their names only'

    def run(self, ctx: ApplicationContext):
        _list_commands(ctx.program_name, ctx.command_registry)


class Application:
    DEFAULT_ENV = 'development'

    def __init__(self, program_name: str,
                 command_class: typing.Optional[typing.Type[Command]] = None,
                 *,
                 description: typing.Optional[str] = None,
                 enable_short_debug_option: bool = False,
                 enable_env_options: bool = True,
                 version: typing.Optional[str] = None
                 ):
        self._program_name = program_name
        self._command_class: typing.Optional[typing.Type[Command]] = command_class
        self._description = description
        self._enable_short_debug_option = enable_short_debug_option if command_class is not None else True
        self._enable_env_options = enable_env_options if command_class is not None else True
        self._version = version
        self._command_registry = CommandRegistry()
        self._command_classes = set()
        self._env_config = EnvConfig(os.environ.get('DEWI_ENV', self.DEFAULT_ENV))
        self._config_dir_registry = ConfigDirRegistry(self._env_config)

        if command_class:
            self._command_classes.add(command_class)
            self._command_registry.register_class(command_class)

    def add_command_class(self, command_class: typing.Type[Command]):
        self._command_classes.add(command_class)
        self._command_registry.register_class(command_class)

    def add_command_classes(self, command_classes: typing.List[typing.Type[Command]]):
        for command_class in command_classes:
            self.add_command_class(command_class)

    def load_plugin(self, name: str):
        self.load_plugins([name])

    def load_plugins(self, names: typing.List[str]):
        try:
            loader = PluginLoader(self._command_registry, self._config_dir_registry)
            loader.load(names)
        except BaseException as exc:
            _print_exception(ApplicationContext(), exc)
            sys.exit(1)

    def register_config_directory(self, directory: str):
        self.register_config_directories([directory])

    def register_config_directories(self, directories: typing.List[str]):
        self._config_dir_registry.register_config_directories(directories)

    def run(self, args: typing.Optional[typing.List[str]] = None):
        single_command_mode = bool(self._command_class and len(self._command_classes) == 1)

        app_context = ApplicationContext()
        app_context.single_command_mode = single_command_mode
        app_context.config_directories = list(self._config_dir_registry.config_directories)
        app_context.environment = self._env_config.current_env

        @click.pass_context
        def app_run(ctx: click.Context, *args, **kwargs):
            app_context.add_cmd_args('__main__', ctx.params, self._command_class.name if single_command_mode else '')
            app_context.command_names.invoked_subcommand = get_invoked_subommand(ctx)
            app_context.command_names.invoked_subcommand_primary_name = ctx.invoked_subcommand
            app_context.current_args = Node.create_from(ctx.params)
            ctx.obj = app_context
            ctx.obj.command_registry = self._command_registry
            ctx.obj.program_name = self._program_name
            for k, v in kwargs.items():
                ctx.obj.add_arg(k, v)
            self._process_debug_opts(ctx.obj.args)
            if self._process_logging_options(ctx.obj.args):
                sys.exit(1)

            if self._enable_env_options and ctx.obj.args.environment:
                self._env_config.set_current_env(ctx.obj.args.environment)

            self._load_env()

            if single_command_mode:
                app_context.command_names.current = self._command_class.name
                res = self._command_class().run(app_context)
                if ctx.invoked_subcommand is None or res:
                    ctx.exit(res)

        app_run.app_context = app_context

        app_opt_ctx = OptionContext()
        app_opt_ctx.register_args(self._register_app_args)
        if single_command_mode:
            app_opt_ctx.register_args(self._command_class.register_arguments)

        try:
            app_run = app_opt_ctx.add_args_to_func(app_run)
        except Exception as e:
            if single_command_mode:
                suffix = f'for command `{self._command_class.name}`'
            else:
                suffix = 'for the application'
            print(f'Error occurred while registering args {suffix}: {e}')
            raise

        if single_command_mode:
            has_classes = len(self._command_class.subcommand_classes) > 0
            app_run = (click.group if has_classes else click.command)(
                self._program_name,
                help=self._description or self._command_class.description,
                context_settings=CONTEXT_SETTINGS,
                cls=(AppGroup if has_classes else AppCommand))(app_run)
            self._register_subcommands(self._command_class.subcommand_classes, app_run, app_context)
        else:
            self._command_registry.register_class(_ListAllCommand)
            self._command_registry.register_class(_ListCommand)

            app_run = click.group(self._program_name, cls=AppGroup,
                                  help=self._description,
                                  context_settings=CONTEXT_SETTINGS,
                                  )(app_run)
            self._register_subcommands(self._command_registry._command_classes, app_run, app_context)

        return app_run(args, self._program_name)

    def _register_subcommands(self, command_classes, parent_run_cmd, app_context: ApplicationContext):
        for command_class in command_classes:
            def wrapper():
                c = command_class

                @click.pass_context
                def r(ctx: click.Context, *_, **kwargs):
                    app_context.add_cmd_args(ctx.info_name, ctx.params)
                    app_context.command_names.current = get_invoked_subommand(ctx.parent)
                    app_context.command_names.invoked_subcommand = get_invoked_subommand(ctx)
                    app_context.command_names.invoked_subcommand_primary_name = ctx.invoked_subcommand
                    app_context.current_args = Node.create_from(ctx.params)
                    for k, v in kwargs.items():
                        ctx.obj.add_arg(k, v)

                    log_debug('Starting command', name=c.name)
                    res = c().run(ctx.obj)
                    if ctx.invoked_subcommand is None or res:
                        ctx.exit(res)

                r.command_class = c

                return r

            try:
                cmd_opt_ctx = OptionContext()
                cmd_opt_ctx.register_args(command_class.register_arguments)
                cls = AliasedGroup if command_class.subcommand_classes else click.core.Command
                cmd_run = parent_run_cmd.command(name=command_class.name, help=command_class.description, cls=cls,
                                                 context_settings=CONTEXT_SETTINGS)(
                    cmd_opt_ctx.add_args_to_func(wrapper()))
                self._register_subcommands(command_class.subcommand_classes, cmd_run, app_context)
            except Exception as e:
                print(f'Error occurred while registering args for command `{command_class.name}`: {e}')
                raise

    def _register_app_args(self, h: OptionContext):
        if self._version:
            h.add_custom_decorator(lambda: click.version_option(self._version))

        h.add_option('--cwd', dest='cwd', help='Change to specified directory')
        h.add_option('--wait', is_flag=True, help='Wait for user input before terminating application')
        h.add_option(
            '--print-backtraces', dest='print_backtraces', is_flag=True,
            help='Print backtraces of the exceptions')

        debug_opts = ['--debug']
        if self._enable_short_debug_option:
            debug_opts.insert(0, '-d')
        h.add_option(*debug_opts, dest='debug', is_flag=True, help='Enable print/log debug messages')
        if self._enable_env_options:
            h.add_option('-e', '--environment', dest='environment',
                         help=f'Specifies the environment to run the app '
                              f'under ({"/".join(sorted(self._env_config.available_envs))})')

        logging = h.add_group('Logging')
        logging.add_option('--log-level', dest='log_level', help='Set log level, default: warning',
                           type=click.Choice([i.name.lower() for i in LogLevel]), default='info')
        logging.add_option('--log-syslog', dest='log_syslog', is_flag=True,
                           help='Log to syslog. Can be combined with other log targets')
        logging.add_option('--log-console', '--log-stdout', dest='log_console', is_flag=True,
                           help='Log to STDOUT, the console. Can be combined with other targets.'
                                'If no target is specified, this is used as default.')
        logging.add_option('--log-file', dest='log_file', multiple=True,
                           help='Log to a file. Can be specified multiple times and '
                                'can be combined with other options.')
        logging.add_option('--no-log', '-l', dest='log_none', is_flag=True,
                           help='Disable logging. If this is set, other targets are invalid.')

    def _process_debug_opts(self, ns: Node):
        if ns.debug or os.environ.get('DEWI_DEBUG', 0) == '1':
            ns.print_backtraces = True
            ns.log_level = 'debug'
            ns.debug_ = True

    def _process_logging_options(self, args: Node):
        return create_logger_from_config(
            LoggerConfig.create(self._program_name, args.log_level, args.log_none, args.log_syslog, args.log_console,
                                args.log_file))

    def _load_env(self):
        self._config_dir_registry.load_env()



# Copyright 2015-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import typing

from dewi_core.command import Command
from dewi_core.loader.context import Context


class Plugin:
    """
    A plugin is an extension of DEWI.
    """

    def get_dependencies(self) -> typing.Iterable[str]:
        return ()

    def load(self, c: Context):
        raise NotImplementedError

    @staticmethod
    def _r(c: Context, t: typing.Type[Command]):
        """
        Registers a Command type into commommandregistry.

        >> from dewi_core.commands.sample import SampleCommand
        >> from dewi_core.context import Context
        >> from dewi_core.loader.plugin import Plugin
        >>
        >>
        >>  class SamplePlugin(Plugin):
        >>      '''Provides "sample" command'''
        >>
        >>      def load(self, c: Context):
        >>          self._r(c, SampleCommand)
        >>
        """
        c.commands.register_class(t)

    @staticmethod
    def _register_config_dir(c: Context, d: str):
        c.config_dir_registry.register_config_directory(d)



# Copyright 2015-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

from dewi_core.commandregistry import CommandRegistry
from dewi_core.config.node import Node
from dewi_core.config_env import ConfigDirRegistry


class ContextError(Exception):
    pass


class ContextEntryNotFound(ContextError):
    pass


class ContextEntryAlreadyRegistered(ContextError):
    pass


class Context(Node):
    """
    A context is a generic purpose registry, which helps
    communicate between different parts of the code. Instead of a global variable
    or module is for storing object(s), the context can be used and passed around
    the necessary objects and functions.

    An example: a CommandRegistry object can be registered into this context.
    """

    RESERVED_KEYS = ['command_registry', 'commands', 'config_dir_registry']

    def __init__(self, command_registry: CommandRegistry, config_dir_registry: ConfigDirRegistry):
        self.command_registry = command_registry
        self.commands = command_registry
        self.config_dir_registry = config_dir_registry

    def register(self, name: str, value):
        """
        Registers an element into the context. It doesn't support overwriting already
        registered entries, in that case `ContextEntryAlreadyRegistered` will be raised.

        :param name: the name of the new entry
        :param value: the value of the new entry
       """
        if name in self:
            raise ContextEntryAlreadyRegistered("Context entry {!r} already registered".format(name))
        self[name] = value

    def unregister(self, name: str):
        """
        Unregisters an already registered entry
        :param name: The name of the entry to be unregistered
        """
        self._check_entry(name)

        if name in self.RESERVED_KEYS:
            raise ContextError(f'Reserved key {name} cannot be unregistered')
        del self.__dict__[name]

    def _check_entry(self, name: str):
        if name not in self:
            raise ContextEntryNotFound("Requested context entry {!r} is not registered".format(name))

    def __getitem__(self, item):
        self._check_entry(item)
        return super().__getitem__(item)

    def __contains__(self, item):
        return item in self.__dict__



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import copy
import typing

from dewi_core.appcontext import ApplicationContext


def serialize_application_context(ctx: ApplicationContext) \
        -> typing.Dict[str, typing.Any]:
    return copy.deepcopy({k: v for k, v in ctx if k != 'command_registry'})


def deserialize_application_context(serialized: typing.Dict[str, typing.Any]) -> ApplicationContext:
    return ApplicationContext.create_from(serialized)



#  Copyright 2022, Laszlo Attila Toth
#  Distributed under the terms of the Apache License, Version 2.0
import datetime


def humanize_time(seconds, format=False, trim_secs=True):
    mins, secs = divmod(seconds, 60)
    hours, mins = divmod(mins, 60)
    if trim_secs:
        secs = int(secs * 100) / 100
    days = 0
    if hours > 24:
        days, hours = divmod(hours, 24)
    if format:
        result = ''
        if days > 0:
            result = '%d days' % (days,)
        result = '%s%02d:%02d:%05.2f' % (result, hours, mins, secs)
        return result
    return days, hours, mins, secs


def localtime():
    return datetime.datetime.now().strftime('%Y%m%d-%H%M%S-%s')



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

from dewi_core.command import Command
from dewi_core.commandplugin import CommandPlugin
from dewi_core.commands.edit.edit import EditCommand
from dewi_core.commands.review import ReviewCommand


class EditOrReviewCommand(Command):
    name = 'edit-review'
    aliases = ['er']
    description = 'Example of command with subcommands from other loaded plugins'
    subcommand_classes = [
        EditCommand,
        ReviewCommand
    ]


EditReviewPlugin = CommandPlugin.create(EditOrReviewCommand)



# Copyright 2019-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0
import os
import typing


def find_file_recursively(filename: str, directory_name: typing.Optional[str] = None) -> typing.Optional[str]:
    """
    Searches a filename in directory_name directory or in current directory
    if directory_name is None, or in their parent directories if the filename file
    is not found in the specific directory.

    The behaviour is the same as how git finds .gitignore, .git/ and other entries.

    :param filename: the filename to be found
    :param directory_name: the start directory, or if it is None: current directory
    :return: the absolute path of the searched file or None if it is not found
    """

    def maybe_file(filename: str, directory_name: str) -> typing.Optional[str]:
        full_path = os.path.join(directory_name, filename)
        if os.path.exists(full_path):
            return full_path
        else:
            return None

    directory_name = os.path.abspath(directory_name) if directory_name else os.getcwd()
    fname = maybe_file(filename, directory_name)
    while fname is None and directory_name and directory_name != os.path.sep:
        directory_name = os.path.dirname(directory_name)
        fname = maybe_file(filename, directory_name)

    return fname



# Copyright 2009-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0
# vim: sts=4 ts=8 et ai

import configparser
import typing


class DictConfigParser(configparser.RawConfigParser):

    def as_dict(self) -> typing.Dict[str, typing.Dict[str, str]]:
        d = dict(self._sections)
        for k in d:
            d[k] = dict(self._defaults, **d[k])
            d[k].pop('__name__', None)
        return d


class IniConfig:

    def __init__(self):
        self.config_file = None
        self.parser = DictConfigParser()
        self.loaded_files = []

    def _section_from_dotted(self, section: str) -> str:
        s = section.split('.', 1)
        if len(s) == 2:
            return '%s "%s"' % tuple(s)
        else:
            return section

    def _section_to_dotted(self, section: str) -> str:
        s = section.split('"')
        if len(s) == 3:
            return '%s.%s' % (s[0][:-1], s[1])
        else:
            return section

    def open(self, cfgfile: str, /, *, merge: bool=False):
        if not merge:
            if self.config_file is not None:
                raise IniConfigError("Already opened with " + self.config_file)
            self.config_file = cfgfile
        elif not self.config_file:
            raise IniConfigError("No initial config is opened")
        self.loaded_files.append(cfgfile)
        self.parser.read(cfgfile, encoding='UTF-8')

    def extend(self, cfgfile: str):
        self.loaded_files.append(cfgfile)
        self.parser.read(cfgfile, encoding='UTF-8')

    def write(self, filename: typing.Optional[str] = None):
        if filename:
            self.config_file = filename
        if self.config_file is None:
            raise IniConfigError("Unable to save config file because file name is unset")
        with open(self.config_file, 'wt') as f:
            self.parser.write(f)

    def close(self):
        self.config_file = None
        self.loaded_files.clear()
        self.parser = DictConfigParser()

    def has(self, section: str, option: str) -> bool:
        section = self._section_from_dotted(section)
        return self.parser.has_option(section, option)

    def set(self, section: str, option: str, value):
        section = self._section_from_dotted(section)
        if not self.parser.has_section(section):
            self.parser.add_section(section)
        self.parser.set(section, option, value)

    def get(self, section: str, option: str) -> typing.Optional[str]:
        section = self._section_from_dotted(section)
        if not self.parser.has_section(section):
            return None
        try:
            return self.parser.get(section, option)
        except configparser.NoOptionError:
            return None

    def get_or_default_value(self, section: str, option: str, default_value: str) -> str:
        res = self.get(section, option)
        return res if res is not None else default_value

    def remove(self, section: str, option: str):
        section = self._section_from_dotted(section)
        if not self.parser.has_section(section):
            return
        if not self.parser.has_option(section, option):
            return
        self.parser.remove_option(section, option)

    def get_options(self, section: str) -> typing.List[str]:
        section = self._section_from_dotted(section)
        if not self.parser.has_section(section):
            return []
        else:
            return self.parser.options(section)

    def get_sections(self) -> typing.List[str]:
        return [self._section_to_dotted(s) for s in self.parser.sections()]

    def as_dict(self) -> typing.Dict[str, typing.Dict[str, str]]:
        return self.parser.as_dict()


class IniConfigError(Exception):
    pass


def convert_to_bool(val) -> bool:
    if not val:
        return False
    if isinstance(val, bool):
        return val
    lowered = val.lower()
    return lowered == 'y' or lowered == 'yes' or lowered == 't' or lowered == 'true'


def convert_from_bool(val: bool) -> str:
    return 'yes' if val else 'no'



# Copyright 2015-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import re
import subprocess

import click

from dewi_core.appcontext import ApplicationContext
from dewi_core.command import Command
from dewi_core.commandplugin import CommandPlugin
from dewi_core.optioncontext import OptionContext


def convert_to_vim_args(args):
    result = []
    if len(args) == 1:
        match = re.match(r'(.*?):([[0-9]+)(:[0-9]+)?:?$', args[0])
        if match:
            result.append(match.group(1))
            result.append('+' + match.group(2))
        else:
            result.append(args[0])
    elif len(args) > 1:
        result = ['-p'] + args

    return result


class EditCommand(Command):
    name = 'edit'
    aliases = ['ed']
    description = 'Calls vim with the file names and line numbers parsed from argument list.'

    @staticmethod
    def register_arguments(c: OptionContext):
        c.add_argument('file_list', nargs=-1, type=click.Path(), help='Files to open')

    def run(self, ctx: ApplicationContext):
        args = ['vim'] + convert_to_vim_args(ctx.args.file_list)
        pipe = subprocess.Popen(args)
        pipe.communicate()


EditPlugin = CommandPlugin.create(EditCommand)



# Copyright 2017-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import collections.abc
import typing

import yaml
from yaml.dumper import Dumper


class Node(collections.abc.MutableMapping):
    """
    This class is a base class to add typesafe objects to a Config.
    Example:

    >>> from dewi_core.config.config import Config
    >>> class A(Node):
    >>>     entry: str = 'default-value'
    >>> c = Config()
    >>> c.set('root', A())
    """

    SEALED_ATTR_NAME = '_sealed__'
    _sealed__ = False

    def _seal(self):
        self._sealed__ = True

    def _unseal(self):
        self._sealed__ = False

    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __setattr__(self, key, value):
        if self._sealed__ and key not in self.__dict__:
            raise KeyError(key)
        super().__setattr__(key, value)

    def __iter__(self):
        return iter({x: y for x, y in self.__dict__.items() if x != self.SEALED_ATTR_NAME})

    def __delitem__(self, key):
        raise RuntimeError('Unable to delete key {}'.format(key))

    def __repr__(self):
        return str({x: y for x, y in self.__dict__.items() if x != self.SEALED_ATTR_NAME})

    def __contains__(self, item):
        return item != self.SEALED_ATTR_NAME and item in self.__dict__

    def load_from(self, data: dict):
        load_node(self, data, sealed=self._sealed__)

    @classmethod
    def create_from(cls, data: dict):
        n = cls()
        n.load_from(data)
        return n


class NodeList(list):
    def __init__(self, member_type: typing.Type[Node]):
        super().__init__()
        self.type_: typing.Type[Node] = member_type

    def load_from(self, data: list):
        self.clear()
        for item in data:
            if isinstance(item, Node):
                self.append(item)
            else:
                node = self.type_()
                node.load_from(item)
                self.append(node)


def load_node(node: Node, d: dict, *, sealed: bool = False):
    for key, value in d.items():
        if key in node and isinstance(node[key], (Node, NodeList)):
            node[key].load_from(value)
        elif key in node or not sealed:
            node[key] = value


def represent_node(dumper: Dumper, data: Node):
    return dumper.represent_dict(data)


def represent_node_list(dumper: Dumper, data: NodeList):
    return dumper.represent_list(data)


yaml.add_multi_representer(Node, represent_node)
yaml.add_multi_representer(NodeList, represent_node_list)



# Copyright 2015-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import typing

from dewi_core.appcontext import ApplicationContext
from dewi_core.optioncontext import OptionContext


class Command:
    name = ''
    aliases = list()
    description = ''
    subcommand_classes = []
    man_page_file: str = ''  # relative to the command's file

    @staticmethod
    def register_arguments(c: OptionContext):
        # can also be @classmethod
        pass

    def run(self, ctx: ApplicationContext) -> typing.Optional[int]:
        pass



# Copyright 2012-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import sys
import typing

from dewi_core.command import Command


class CommandRegistryException(Exception):
    pass


class ClassNotFound(CommandRegistryException):
    pass


class ClassIsNotSubclassOfCommand(Exception):
    pass


class ClassDescriptor:
    """
    Describes how to create a class based on specific values in subclasses.

    There is also a method that lets the class to be get - without instantiation of it.
    """

    def get_class(self) -> typing.Type[Command]:
        raise NotImplementedError()

    def get_name(self):
        raise NotImplementedError()


class ClassDescriptorWithModuleAndClassName(ClassDescriptor):
    """
    Both the module name and the command class name are specified
    """

    def __init__(self, module_name: str, class_name: str):
        super().__init__()
        self._module_name = module_name
        self._class_name = class_name

    def get_class(self) -> typing.Type[Command]:
        module = self._get_module_object()
        try:
            return getattr(module, self._class_name)
        except AttributeError:
            raise ClassNotFound("Invalid class or attribute: {0}".format(self._class_name))

    def _get_module_object(self):
        __import__(self._module_name)
        return sys.modules[self._module_name]

    def get_name(self) -> str:
        return self.get_class().name


class ClassDescriptorWithModuleName(ClassDescriptorWithModuleAndClassName):
    """The module and the class has the same name (the module's last part at least)"""

    def __init__(self, module_name: str):
        super().__init__(module_name, module_name.split('.')[-1])

    def get_name(self) -> str:
        return self.get_class().name


class ClassDescriptorWithModuleNameAndCommandClassMember(ClassDescriptorWithModuleAndClassName):
    """
    The command is in a module that have specific layout, its __init__.py has commandlcass member. This member
    stores the command class.
    """

    def __init__(self, module_name: str):
        super().__init__(module_name, 'command_class')


class ClassDescriptorWithConcreteClass(ClassDescriptor):
    """
    This descriptor stores the class itself, mainly used for testing.
    """

    def __init__(self, class_object: typing.Type[Command]):
        if not issubclass(class_object, Command):
            raise ClassIsNotSubclassOfCommand(
                'The {} class is not subclass of dewi_core.command.Command'.format(class_object.__name__))
        super().__init__()
        self.class_object = class_object

    def get_class(self) -> typing.Type[Command]:
        return self.class_object

    def get_name(self) -> str:
        return self.class_object.name


class CommandRegistry:
    '''
    Registry of command classes
    '''

    def __init__(self):
        self._registry = dict()
        self._command_classes: typing.List[typing.Type[Command]] = []

    def _validate_name_and_class_descriptor(self, name: str, class_descriptor: ClassDescriptor):
        if name in self._registry:
            raise CommandRegistryException("Command class name is already used; name='{0}'".format(name))
        if not isinstance(class_descriptor, ClassDescriptor):
            raise CommandRegistryException(
                "Command class descriptor's type differs from ClassDescriptor; "
                "type='{0}'".format(type(class_descriptor).__name__))

    def register_command_class(self, name: str, class_descriptor: ClassDescriptor):
        self._validate_name_and_class_descriptor(name, class_descriptor)
        self._registry[name] = class_descriptor

    def register_class(self, command_class: typing.Type[Command]):
        desc = ClassDescriptorWithConcreteClass(command_class)
        self.register_command_class(command_class.name, desc)
        for alias in command_class.aliases:
            self.register_command_class(alias, desc)
        self._command_classes.append(command_class)

    def get_command_class_descriptor(self, name: str) -> ClassDescriptor:
        if name not in self._registry:
            raise CommandRegistryException("Specified command class name is not found; name='{0}'".format(name))
        return self._registry[name]

    def get_command_count(self) -> int:
        return len(self._registry)

    def get_command_names(self) -> typing.List[str]:
        return list(self._registry.keys())

    def __contains__(self, command_name: str) -> bool:
        return command_name in self._registry



import os
import os.path
import pathlib
import typing

from dewi_core.config.appconfig import get_config as app_config
from dewi_core.config.iniconfig import IniConfig
from dewi_core.logger import log_debug
from dewi_core.utils.files import find_file_recursively

ANY_VERSION = '*'
INITIAL_VERSION = '0'


class Project:
    CFG_FILE = 'project.cfg'
    VERSION = 2

    def __init__(self, name: str):
        self.name = name
        self._project_dir = f'{app_config().projectdir}/{name}'
        self.config = IniConfig()
        self.config.open(f'{self._project_dir}/{self.CFG_FILE}')
        self.ext = _project_ext_type(self) if _project_ext_type else None

    @property
    def project_dir(self) -> str:
        return self._project_dir

    @property
    def etc_dir(self) -> str:
        return f'{self._project_dir}/etc'

    @property
    def src_dir(self) -> str:
        return f'{self._project_dir}/src'

    @property
    def tmp_dir(self) -> str:
        return f'{self._project_dir}/tmp'

    @property
    def branch(self) -> str:
        return self.config.get('main', 'branch')

    @property
    def upstream_branch(self) -> str:
        return self.config.get('main', 'upstream_branch')

    @property
    def remote(self) -> str:
        return self.config.get('main', 'remote')

    @property
    def ticket_id(self) -> str:
        return self.config.get('main', 'ticket_id')

    def repo_dir(self, repo: str) -> str:
        return f'{self.src_dir}/{repo}'

    @property
    def repositories(self) -> typing.Iterable[str]:
        for entry in os.listdir(self.src_dir):
            if os.path.exists(f'{self.repo_dir(entry)}/.git'):
                yield entry

    @property
    def version(self) -> str:
        return self.config.get('core', 'version')

    @property
    def custom_version(self) -> str:
        return self.config.get_or_default_value('core', 'custom_version', INITIAL_VERSION)

    def as_dict(self) -> typing.Dict[str, typing.Any]:
        return dict(
            name=self.name,
            branch=self.branch,
            upstream=dict(remote=self.remote,
                          branch=self.upstream_branch),
            ticket_id=self.ticket_id,
            repositories=list(self.repositories)
        )


class ProjectExtension:
    """
    Additional fields and functionality to extend Project class without changing it.
    """

    def __init__(self, project: Project):
        self.project = project
        self.config = project.config


class ProjectError(RuntimeError):
    pass


class ProjectUpdateError(ProjectError):
    pass


def create_project(name: typing.Optional[str] = None) -> Project:
    return _create_project(name)


def _create_project(name: typing.Optional[str] = None, *,
                    branch: typing.Optional[str] = None,
                    upstream_remote: typing.Optional[str] = None,
                    upstream_branch: typing.Optional[str] = None,
                    ticket_id: typing.Optional[str] = None) -> Project:
    project_dir = app_config().projectdir
    if not name or name == '.':
        fname = find_file_recursively(Project.CFG_FILE)
        if not fname and os.getcwd().startswith(project_dir):
            fname = os.path.join(project_dir,
                                 os.getcwd()[len(project_dir) + len(os.sep):].split(os.sep, 1)[0],
                                 Project.CFG_FILE)
        if not fname:
            raise ProjectError(f"Current directory is not part of any project; directory={os.getcwd()}")
        name = os.path.basename(os.path.dirname(fname))

    log_debug('Creating project directory if not exists', dict(directory=os.path.join(project_dir, name)))
    os.makedirs(os.path.join(project_dir, name), exist_ok=True)
    update_project(name, branch=branch, upstream_remote=upstream_remote, upstream_branch=upstream_branch,
                   ticket_id=ticket_id)
    return Project(name)


def create_project_with_details(name: str, *, branch: str, upstream_remote: str, upstream_branch: str,
                                ticket_id: str) -> Project:
    return _create_project(name, branch=branch, upstream_remote=upstream_remote, upstream_branch=upstream_branch,
                           ticket_id=ticket_id)


def update_project(project_name: str, *,
                   branch: typing.Optional[str] = None,
                   upstream_remote: typing.Optional[str] = None,
                   upstream_branch: typing.Optional[str] = None,
                   ticket_id: typing.Optional[str] = None):
    project_dir = f'{app_config().projectdir}/{project_name}'
    if not os.path.exists(os.path.join(project_dir, Project.CFG_FILE)):
        log_debug('Initial, still non-existing project config')
        core_version = INITIAL_VERSION
        custom_version = INITIAL_VERSION
    else:
        log_debug('Selecting version specific Updater')
        core_version, custom_version = _get_core_and_custom_version(project_dir)

    while True:
        updaters = []
        if core_version in _registered_core_updaters:
            updaters += _registered_core_updaters[core_version]
        updaters += _registered_core_updaters[ANY_VERSION]

        if updaters:
            MultiUpdater(project_dir, branch=branch, upstream_remote=upstream_remote, upstream_branch=upstream_branch,
                         ticket_id=ticket_id,
                         updaters=updaters).update()

        while True:
            updaters = []
            if custom_version in _registered_updaters:
                updaters += _registered_updaters[custom_version]
            updaters += _registered_updaters[ANY_VERSION]

            if not updaters:
                break

            MultiUpdater(project_dir, branch=branch, upstream_remote=upstream_remote, upstream_branch=upstream_branch,
                         ticket_id=ticket_id,
                         updaters=updaters).update()

            _, new_custom_version = _get_core_and_custom_version(project_dir)
            if new_custom_version == custom_version:
                break
            else:
                custom_version = new_custom_version

        new_core_version, _ = _get_core_and_custom_version(project_dir)

        if new_core_version == core_version:
            break

        else:
            core_version = new_core_version


def _get_core_and_custom_version(project_dir: str) -> typing.Tuple[str, str]:
    cfg = IniConfig()
    cfg.open(os.path.join(project_dir, Project.CFG_FILE))
    core_version = cfg.get('core', 'version')
    custom_version = cfg.get_or_default_value('core', 'custom_version', INITIAL_VERSION)
    return core_version, custom_version


class Updater:
    def __init__(self, project_dir: str, ):
        self._project_dir = project_dir

    def update(self):
        raise NotImplementedError()

    def _create_dir_if_missing(self, directory: str, *, with_dot_keep_file: bool = False):
        directory = os.path.join(self._project_dir, directory)
        if not os.path.exists(directory):
            log_debug('Creating project subdirectory', dict(directory=directory))
            os.makedirs(directory, exist_ok=True)
        if with_dot_keep_file:
            filename = os.path.join(directory, '.keep')
            if not os.path.exists(filename):
                log_debug('Add missing .keep file', dict(directory=directory))
                pathlib.Path(filename).touch()

    def _open_cfg(self) -> IniConfig:
        cfg = IniConfig()
        cfg.open(os.path.join(self._project_dir, Project.CFG_FILE))
        return cfg

    def _core_version(self, cfg: IniConfig) -> str:
        return cfg.get('core', 'version')

    def _custom_version(self, cfg: IniConfig) -> str:
        return cfg.get_or_default_value('core', 'custom_version', INITIAL_VERSION)

    def _set_custom_version(self, cfg: IniConfig, version: str):
        cfg.set('core', 'custom_version', version)


class DetailedUpdater(Updater):
    def __init__(self, project_dir: str, *,
                 branch: typing.Optional[str] = None,
                 upstream_remote: typing.Optional[str] = None,
                 upstream_branch: typing.Optional[str] = None,
                 ticket_id: typing.Optional[str] = None):
        super().__init__(project_dir)
        self._name = os.path.basename(self._project_dir)
        self._branch = branch or ''
        self._upstream_remote = upstream_remote or ''
        self._upstream_branch = upstream_branch or ''
        self._ticket_id = ticket_id or ''


class NoopUpdater(Updater):
    def __init__(self, project_dir: str):
        super().__init__(project_dir)

    def update(self):
        pass


class MultiUpdater(DetailedUpdater):
    def __init__(self, project_dir: str, *,
                 branch: typing.Optional[str] = None,
                 upstream_remote: typing.Optional[str] = None,
                 upstream_branch: typing.Optional[str] = None,
                 ticket_id: typing.Optional[str] = None,
                 updaters: typing.List[typing.Type[Updater]]):
        super().__init__(project_dir, branch=branch, upstream_remote=upstream_remote, upstream_branch=upstream_branch,
                         ticket_id=ticket_id)
        self._updaters = updaters

    def update(self):
        for updater_cls in self._updaters:
            if issubclass(updater_cls, DetailedUpdater):
                updater = updater_cls(self._project_dir, branch=self._branch, upstream_remote=self._upstream_remote,
                                      upstream_branch=self._upstream_branch,
                                      ticket_id=self._ticket_id)
            else:
                updater = updater_cls(self._project_dir)

            updater.update()


class ProjectUpdaterV1(DetailedUpdater):

    def update(self):
        self._ensure_subdir_hierarchy()
        self._write_project_config()

    def _ensure_subdir_hierarchy(self):
        self._create_dir_if_missing('src', with_dot_keep_file=True)
        self._create_dir_if_missing('etc', with_dot_keep_file=True)
        self._create_dir_if_missing('tmp', with_dot_keep_file=True)

    def _write_project_config(self):
        if not self._ticket_id or not self._upstream_branch or not self._branch:
            raise ProjectUpdateError('Missing details')
        cfg = IniConfig()
        cfg.config_file = os.path.join(self._project_dir, Project.CFG_FILE)
        cfg.set('core', 'version', '1')
        cfg.set('core', 'custom_version', INITIAL_VERSION)
        # branch is git branch, name is the project directory name
        cfg.set('core', 'name', self._name)
        cfg.set('main', 'branch', self._branch)
        cfg.set('main', 'ticket_id', self._ticket_id)
        cfg.set('main', 'maint_branch', self._upstream_branch)
        cfg.set('main', 'remote', self._upstream_remote)
        cfg.write()


class ProjectUpdaterV2(Updater):

    def update(self):
        cfg = self._open_cfg()
        cfg.set('core', 'version', '2')
        cfg.set('main', 'upstream_branch', cfg.get('main', 'maint_branch'))
        cfg.remove('main', 'maint_branch')
        cfg.write()


# Each version can have one or more updater
# '0' means: initial update
# '*' means: always run regardless of current version, aka. mandatory updater
# Behaviour:
#   for every core version starting with zero:
#        run version specific core updaters
#        and run mandatory core updaters (ANY_VERSION)
#        for every custom version starting with current version (initially: zero)
#                    run version specific custom updaters
#                    run mandatory custom updaters
# It is possible that later the actul upgrade will be the following sequence:
#   core updater to v1, custom updater to v1, core updater to v2, custom to v2 and v3, core to v3, custom to v4 etc.
# as each update can check the project's current state at that step and they may depend on a previous updater

_registered_core_updaters: typing.Dict[str, typing.List[typing.Type[Updater]]] = {
    ANY_VERSION: [],
    INITIAL_VERSION: [ProjectUpdaterV1],
    '1': [ProjectUpdaterV2],
}

# custom updaters based on core.custom_version field
_registered_updaters: typing.Dict[str, typing.List[typing.Type[Updater]]] = {ANY_VERSION: [], INITIAL_VERSION: []}

_project_ext_type: typing.Optional[typing.Type[ProjectExtension]] = None


def register_updater(version: str, updater: typing.Type[Updater]):
    global _registered_updaters
    if version not in _registered_updaters:
        _registered_updaters[version] = []

    _registered_updaters[version].append(updater)


def register_updater_to_all(updater: typing.Type[Updater]):
    _registered_updaters[ANY_VERSION].append(updater)


def register_project_extension(ext: typing.Type[ProjectExtension]):
    global _project_ext_type

    _project_ext_type = ext



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

from dewi_core.command import Command
from dewi_core.commandplugin import CommandPlugin
from .commands import GerritChangeReviewer, GerritChangeChainReviewer, register_global_args_in_review_cmd
from ...optioncontext import OptionContext


class ReviewCommand(Command):
    name = 'review'
    aliases = ['r', 'rw']
    description = "Example command which could be for Gerrit review"
    subcommand_classes = [
        GerritChangeChainReviewer,
        GerritChangeReviewer,
    ]

    @classmethod
    def register_arguments(cls, c: OptionContext):
        register_global_args_in_review_cmd(c)


ReviewPlugin = CommandPlugin.create(ReviewCommand)



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import typing

from dewi_core.config.node import Node


class RunningCommandNames(Node):
    def __init__(self):
        # the name how current Command is called (which alias)
        self.current: str = ''
        # for the currently running command the invoked subcommand (eg. in the run() of ReviewCommand may be 'change')
        self.invoked_subcommand: typing.Optional[str] = None
        self.invoked_subcommand_primary_name: typing.Optional[str] = None
        # top-level command name, in single command mode it's that command, otherwise first subcommand
        self.command: str = ''
        # both running_command and running_subcommands
        self.subcommands: typing.List[str] = []
        # both command and subcommands
        self.running_full_command: typing.List[str] = []
        self._seal()


class ApplicationContext(Node):
    def __init__(self):
        from .commandregistry import CommandRegistry
        # args: similar to argparse.Namespace: a single mapping containing all options.
        # It may have conflicting keys, so latest subcommand wins
        self.args = Node()
        # Each command has its own args (options), this contains them separately
        # The app (top-level cmd) belongs to the '__main__' key, the first-level subcommand
        # into the 'subcmd-name', second-level: 'subcmd-name subcmd2-name2'
        self.commands_args = Node()
        self.current_args = Node()
        self.program_name = ''
        self.command_registry: CommandRegistry = None
        self.single_command_mode: bool = False
        self.command_names = RunningCommandNames()
        self.config_directories: typing.List[str] = []
        self.environment: str = ''
        # A generic context for subcommands, may pass data to their subcommands
        self.command_context = Node()
        self._seal()

    def add_arg(self, key: str, value):
        setattr(self.args, key, value)

    def add_cmd_args(self, cmd: str, args: typing.Dict, single_cmd_name=None):
        if cmd == '__main__':
            self.commands_args[cmd] = Node.create_from(args)
        if single_cmd_name:
            cmd = single_cmd_name
        if cmd != '__main__':
            if not self.command_names.command:
                self.command_names.command = cmd
            else:
                self.command_names.subcommands.append(cmd)
            self.command_names.running_full_command.append(cmd)

            self.commands_args[' '.join(self.command_names.running_full_command)] = Node.create_from(args)



# vim: ts=8 sts=4 sw=4 et ai si
import abc
import argparse
import inspect
import os
import random
import re
import subprocess
import typing

from dewi_core.command import Command

XMLTO_XSL_DIR = os.path.join(os.path.dirname(__file__), 'xmlto_xsl')


class GeneratorConfig:
    def __init__(self, includes_dir: str, output_dir: str, app_name: str, *, app_version: typing.Optional[str] = None,
                 section: int = 1):
        self.includes_dir = includes_dir
        self.output_dir = output_dir
        self.app_name = app_name
        self.app_version = app_version or '1.0-SNAPSHOT'
        self.section = section

    def get_basename(self, cmd_name: str):
        if cmd_name:
            return f'{self.app_name}-{cmd_name.replace(" ", "-")}'
        else:
            return self.app_name

    def get_refentry_title(self, cmd_name: str):
        if cmd_name:
            return f'{self.app_name} {cmd_name}'
        else:
            return self.app_name

    def get_target_filename(self, cmd_name: str):
        return f'{self.output_dir}/{self.get_basename(cmd_name)}.{self.section}'

    def get_include_filename(self, include_name: str):
        return f'{self.includes_dir}/{include_name}.xml'


class Generator:
    def __init__(self, config: GeneratorConfig):
        self.config = config
        self._random = random.randint(1000, 100000)

    def _get_temp_dir(self):
        return f'/tmp/DEWI-MAN-{self._random}-{os.getpid()}/man-tmp'

    def _get_temp_file(self, cmd_name: str):
        return self._get_temp_dir() + '/tmp--' + self.config.get_basename(cmd_name) + '.xml.' + str(os.getpid())

    def print_man_page(self, source_file: str, cmd_name: str, doc: str):
        try:
            self.generate_man_page(source_file, cmd_name)
            if os.path.exists(self.config.get_target_filename(cmd_name)):
                subprocess.run(['man', self.config.get_target_filename(cmd_name)])
            else:
                print(doc)
        except subprocess.CalledProcessError:
            print("\nUnable to generate manual page. Perhaps \"xmlto\" is not installed?\n")
            print(doc)

    def generate_man_page(self, source_file: str, cmd_name: str) -> bool:
        if os.path.exists(source_file):
            self._generate_man_page(source_file)
            return True
        else:
            return False

    def _generate_man_page(self, source_file: str, cmd_name: str):
        if self._is_update_of_man_page_needed(source_file, cmd_name):
            self._update_man_page(source_file, cmd_name)

    def _is_update_of_man_page_needed(self, source_file: str, cmd_name: str):
        man_file = self.config.get_target_filename(cmd_name)
        xml_file = source_file

        def is_newer(f1: str, f2: str) -> bool:
            return os.path.getmtime(f1) > os.path.getmtime(f2)

        update = (not os.path.exists(man_file) or is_newer(xml_file, man_file))

        if not update:
            xf_content = open(xml_file).read().split('\n')
            for xf in xf_content:
                if re.match(r'<!-- include [^ ]+ -->', xf):
                    fname = self.config.get_include_filename(re.sub(r'^<!-- include ([^ ]+) -->', r'\1', xf))
                    if is_newer(fname, man_file):
                        update = True
                        break

        return update

    def _update_man_page(self, source_file: str, cmd_name: str):
        converted = ''
        refentrytitle_changed = False
        xf_content = open(source_file, encoding='UTF-8').read().split('\n')
        for xf in xf_content:
            if re.match(r'.*<refmiscinfo class="version">.*</refmiscinfo>', xf):
                converted += re.sub(
                    r'(<refmiscinfo class="version">).*(</refmiscinfo>)',
                    f'<refmiscinfo class="version">{self.config.app_version}</refmiscinfo>\n', xf)
            elif re.match(r'.*<refname>.*</refname>', xf):
                converted += '<refname>{0}</refname>\n'.format(os.path.basename(source_file)[:-4])
            elif not refentrytitle_changed and re.match(r'.*<refentrytitle>.*</refentrytitle>', xf):
                converted += f'<refentrytitle>{self.config.get_refentry_title()}</refentrytitle>\n'
                refentrytitle_changed = True
            elif re.match(r'<!-- include [^ ]+ -->', xf):
                fname = self.config.get_include_filename(re.sub(r'^<!-- include ([^ ]+) -->', r'\1', xf))
                converted += open(fname).read()
            elif re.match(r'.*\$CMDNAME\b.*', xf):
                converted += '%s\n' % re.sub(r'\$CMDNAME\b', self.config.app_name, xf)
            else:
                converted += '%s\n' % (xf,)

        os.makedirs(self._get_temp_dir(), exist_ok=True)
        tempfile = open(self._get_temp_file(cmd_name), "w")
        tempfile.write(converted)
        tempfile.close()

        os.makedirs(self.config.output_dir, exist_ok=True)

        subprocess.run([
            'xmlto', '-o', self.config.output_dir, '-m', 'manpage-normal.xsl',
            '-m', 'manpage-bold-literal.xsl', 'man', self._get_temp_file(cmd_name)],
            cwd=XMLTO_XSL_DIR, check=True)


class ManGeneratorConfigBase(abc.ABC):
    def __init__(self, app_name: str, man_name: str):
        self._app_name = app_name
        self._name = man_name
        self.tmpfile = ''
        self._random = random.randint(1000, 100000)

    def _set_name(self, name: str):
        self._name = name

    def get_name(self):
        return self._name

    @abc.abstractmethod
    def get_source_dir(self) -> str:
        pass

    @abc.abstractmethod
    def get_target_dir(self) -> str:
        pass

    def get_source_file(self):
        return self.get_source_dir() + '/' + self._get_base_file_name() + '.xml'

    def get_target_file(self):
        return f'{self.get_target_dir()}/{self._get_target_base_file_name()}.{self._get_section()}'

    def _get_section(self):
        return 1

    def get_temp_dir(self):
        return f'/tmp/DEWI-MAN-{self._random}-{os.getpid()}/man-tmp'

    def get_temp_file(self):
        if not self.tmpfile:
            self.tmpfile = self.get_temp_dir() + '/tmp--' + self._get_base_file_name() + '.xml.' + str(os.getpid())
        return self.tmpfile

    def _get_base_file_name(self):
        if self._name:
            return f'{self._app_name}-{self._name.replace(" ", "-")}'
        else:
            return self._app_name

    def _get_target_base_file_name(self):
        return self._get_base_file_name()

    def get_man_page_refname(self):
        return self._get_target_base_file_name()

    def get_man_page_refentrytitle(self):
        if self._name:
            return f'{self._app_name} {self._name}'
        else:
            return self._app_name

    def get_command_name(self):
        return self._app_name


def print_man_page(config: GeneratorConfig, ns: argparse.Namespace):
    cmd_class: typing.Type[Command] = ns.cmd_class_
    man_page_files = [(inspect.getfile(cmd_class), cmd_class.man_page_file)]
    name = ns.running_command_
    if 'running_subcommands_' in ns and ns.running_subcommands_:
        for sub_cmd_name in ns.running_subcommands_:
            for cc in cmd_class.subcommand_classes:
                if cc.name == sub_cmd_name:
                    cmd_class = cc
                    man_page_files.append((inspect.getfile(cmd_class), cmd_class.man_page_file))
                    name += f' {sub_cmd_name}'

    man_page_file = ''
    for entry in reversed(man_page_files):
        if entry[1]:
            man_page_file = os.path.abspath(os.path.join(os.path.dirname(entry[0]), entry[1]))

    g = Generator(config)
    # a subcommand may not have separate man page file but its use is mandatory
    # if no man page is specified (and the current method is called)
    return g.print_man_page(man_page_file, name, cmd_class.__doc__)



# Copyright 2018-2021 Tóth, László Attila
# Distributed under the terms of the Apache License, Version 2.0

# R0913: too many argument
# R0914: too many locals (get_similar_names_to)
# pylint: disable=R0913,R0914


def qsort(lst, func=None):
    if len(lst) <= 1:
        return lst
    if func is None:
        func = type(lst[0]).__lt__
    pivot = lst.pop(0)
    greater_eq = qsort([i for i in lst if not func(i, pivot)], func)
    lesser = qsort([i for i in lst if func(i, pivot)], func)
    return lesser + [pivot] + greater_eq


# from git's levenshtein.c:
#
# This function implements the Damerau-Levenshtein algorithm to
# calculate a distance between strings.
#
# Basically, it says how many letters need to be swapped, substituted,
# deleted from, or added to string1, at least, to get string2.
#
# The idea is to build a distance matrix for the substrings of both
# strings.  To avoid a large space complexity, only the last three rows
# are kept in memory (if swaps had the same or higher cost as one deletion
# plus one insertion, only two rows would be needed).
#
# At any stage, "i + 1" denotes the length of the current substring of
# string1 that the distance is calculated for.
#
# row2 holds the current row, row1 the previous row (i.e. for the substring
# of string1 of length "i"), and row0 the row before that.
#
# In other words, at the start of the big loop, row2[j + 1] contains the
# Damerau-Levenshtein distance between the substring of string1 of length
# "i" and the substring of string2 of length "j + 1".
#
# All the big loop does is determine the partial minimum-cost paths.
#
# It does so by calculating the costs of the path ending in characters
# i (in string1) and j (in string2), respectively, given that the last
# operation is a substition, a swap, a deletion, or an insertion.
#
# This implementation allows the costs to be weighted:
#
# - w (as in "sWap")
# - s (as in "Substitution")
# - a (for insertion, AKA "Add")
# - d (as in "Deletion")
#
# Note that this algorithm calculates a distance _iff_ d == a.
#
def levenshtein(str1, str2, w, a, s, d):
    len1 = len(str1)
    len2 = len(str2)
    row0 = [0] * (len2 + 1)
    row1 = [0] * (len2 + 1)
    row2 = [0] * (len2 + 1)

    for i in range(0, len2 + 1):
        row1[i] = i * a

    for i in range(0, len1):
        row2[0] = (i + 1) * d

        for j in range(0, len2):
            # subst
            if str1[i] != str2[j]:
                row2[j + 1] = row1[j] + s
            else:
                row2[j + 1] = row1[j]

            # swap
            if i > 0 and j > 0 and str1[i - 1] == str2[j] \
                    and str1[i] == str2[j - 1] \
                    and row2[j + 1] > row0[j - 1] + w:
                row2[j + 1] = row0[j - 1] + w

            # deletion
            if row2[j + 1] > row1[j + 1] + d:
                row2[j + 1] = row1[j + 1] + d

            # insertion
            if row2[j + 1] > row2[j] + a:
                row2[j + 1] = row2[j] + a

        dummy = list(row0)
        row0 = row1
        row1 = row2
        row2 = dummy

    return row1[len2]


def get_similar_names_to(name, possible_names, max_result_count=6, w=0, a=1, s=2, d=3):
    # values of w, a, s, d are from git's commit c41494f8c89c330cc9bd4629cdc3c428803e76bc
    if not possible_names:
        return []

    distances = dict()
    for current_name in possible_names:
        distances[current_name] = levenshtein(name, current_name, w, a, s, d)

    def levenshtein_compare(s1, s2):
        nonlocal distances
        l1 = distances[s1]
        l2 = distances[s2]

        if l1 != l2:
            return l1 <= l2

        return s1 <= s2

    sorted_names = qsort(possible_names, levenshtein_compare)

    best_similarity = distances[sorted_names[0]]
    count = 1
    name_len = len(sorted_names)
    while count < name_len and best_similarity == distances[sorted_names[count]]:
        count = count + 1

    similar_names = list()
    if count < max_result_count:
        for i in range(count):
            similar_names.append(sorted_names[i])
    return similar_names



# Copyright 2016-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0
import io
import os
import sys


def redirect_outputs(stdout=None, stderr=None):
    class Redirection:
        def __init__(self):
            self.stdout = stdout or io.StringIO()
            self.stderr = stderr or io.StringIO()

            self._saved_stdouts = []
            self._saved_stderrs = []

        def __enter__(self):
            self._saved_stdouts.append(sys.stdout)
            self._saved_stderrs.append(sys.stderr)

            sys.stdout = self.stdout
            sys.stderr = self.stderr

            return self

        def __exit__(self, *args):
            sys.stdout = self._saved_stdouts.pop()
            sys.stderr = self._saved_stderrs.pop()

    return Redirection()


def in_directory(directory: str):
    class InDirectory:
        def __init__(self, directory: str):
            self._directory = directory
            self._old_directories = list()

        def __enter__(self):
            self._old_directories.append(os.getcwd())
            os.chdir(self._directory)

        def __exit__(self, exc_type, exc_val, exc_tb):
            os.chdir(self._old_directories.pop())

    return InDirectory(directory)



# Copyright 2016-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import typing

import yaml


class InvalidEntry(KeyError):
    pass


class Config:
    def __init__(self):
        self._config = dict()

    def get_config(self):
        return dict(self._config)

    def overwrite_config(self, config: dict):
        self._config = config

    def _top_level_unsafe_set(self, key: str, value):
        self._config[key] = value

    def set(self, entry: str, value):
        cfg, key = self._get_container_and_key(entry)

        if key in cfg and isinstance(cfg[key], (dict, list, set)):
            raise InvalidEntry("The '{}' entry should not refer either a dict, list or set".format(entry))

        cfg[key] = value

    def _get_container_and_key(self, entry):
        parents = entry.split('.')
        key = parents.pop()
        cfg = self._config

        for parent in parents:
            if parent not in cfg:
                cfg[parent] = dict()

            cfg = cfg[parent]

        return cfg, key

    def append(self, list_entry: str, value):
        cfg, key = self._get_container_and_key(list_entry)

        if key not in cfg:
            cfg[key] = list()

        cfg[key].append(value)

    def add_to_set(self, list_entry: str, value):
        cfg, key = self._get_container_and_key(list_entry)

        if key not in cfg:
            cfg[key] = set()

        cfg[key].add(value)

    def get(self, entry: str):
        keys = entry.split('.')
        cfg = self._config

        try:
            for key in keys:
                cfg = cfg[key]
        except KeyError:
            return None

        return cfg

    def delete(self, list_entry: str):
        cfg, key = self._get_container_and_key(list_entry)
        del cfg[key]

    def dump(self, file, ignore: typing.Optional[typing.List[str]] = None):
        cfg = self._config
        if ignore:
            cfg = dict(cfg)
            for item in ignore:
                del cfg[item]

        yaml.dump(cfg, file, indent=4, default_flow_style=False)

    def print(self, file=None):
        self._print(self._config, '', file=file)

    def _print(self, config: dict, prefix: str, file=None):
        for key in sorted(config.keys()):
            value = config[key]

            if isinstance(value, dict):
                self._print(value, "{}{}.".format(prefix, key), file=file)
            else:
                print("{}{}: {}".format(prefix, key, value), file=file)



# Copyright 2020-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import os
import sys

from dewi_core.application import Application
from dewi_core.commands.edit.edit import EditCommand


def main():
    if int(os.environ.get('DEWI_CORE_DEV_WITH_PLUGINS', '0')) == 1:
        app = Application('dewi-edit')
        app.load_plugin('dewi_core.commands.edit.edit.EditPlugin')
    else:
        app = Application('dewi-edit', EditCommand)
    app.run(sys.argv[1:])


if __name__ == '__main__':
    main()



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import importlib.util
import os
import typing

from dewi_core.logger import log_debug


class EnvConfig:
    def __init__(self, current_env: str):
        self._current_env = current_env
        self.available_envs = {'development', 'production'}
        self._config_dir_to_envs_map: typing.Dict[str, typing.Set[str]] = {}

    @property
    def current_env(self):
        return self._current_env

    def set_current_env(self, current_env: str):
        self._current_env = current_env

    def add_from_directory(self, directory: str):
        self._config_dir_to_envs_map[directory] = set()
        envs = set()
        for dir_entry in os.listdir(directory):
            entry = f'{directory}/{dir_entry}'
            if dir_entry == 'environments' and os.path.isdir(entry):
                for env_entry in os.listdir(entry):
                    if env_entry.endswith('.py') and env_entry != '__init__.py':
                        envs.add(env_entry[:-3])

        self.available_envs |= envs
        self._config_dir_to_envs_map[directory] |= envs

    def load_current_env_from_directory(self, directory: str):
        path = os.path.abspath(os.path.join(directory, 'environment.py'))
        if os.path.exists(path):
            self._load(path)

        if self._current_env in self._config_dir_to_envs_map[directory]:
            path = os.path.abspath(os.path.join(directory, 'environments', f'{self._current_env}.py'))
            self._load(path)

    def _load(self, path: str):
        module = os.path.basename(path)[:-3].replace('/', '.').replace('\\', '.')
        log_debug('Loading environment file', dict(filename=path, module_name=module, current_env=self._current_env))
        spec = importlib.util.spec_from_file_location(module, path)
        mfspec = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mfspec)


class ConfigDirRegistry:
    def __init__(self, env_config: EnvConfig):
        self.env_config = env_config
        self.config_directories: typing.List[str] = []

    def register_config_directory(self, directory: str):
        if directory in self.config_directories:
            log_debug('Not registering duplicated config directory', dict(name=directory))
            return False
        log_debug('Registering config directory', dict(name=directory))
        self.config_directories.append(directory)
        self.env_config.add_from_directory(directory)
        return True

    def register_config_directories(self, directories: typing.List[str]):
        for d in directories:
            self.register_config_directory(d)

    def set_current_env(self, env_name: str):
        self.env_config.set_current_env(env_name)

    def load_env(self):
        for config_dir in self.config_directories:
            self.env_config.load_current_env_from_directory(config_dir)


def load_config_of_env(config_dir: str, env_name: str):
    """
    Load a single env file with a common one if exits from the specified directory.
    Useful for loading test env as Rails-like loading is not supported in Python.
    """
    r = ConfigDirRegistry(EnvConfig(env_name))
    r.register_config_directory(config_dir)
    r.load_env()



# Copyright 2018-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import os
import sys
import typing

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


if os.environ.get('DEWI_YAML_WITH_ALIASES', '0') == '0':
    yaml.Dumper.ignore_aliases = lambda *args: True


def save_to_yaml(cfg, output_file: typing.Optional[str] = None):
    if not output_file or output_file == '-':
        yaml.dump(cfg, stream=sys.stdout, indent=4, default_flow_style=False)
    else:
        with open(output_file, 'wt', encoding='UTF-8') as f:
            yaml.dump(cfg, stream=f, indent=4, default_flow_style=False)


def print_as_yaml(cfg):
    save_to_yaml(cfg, '-')


def load_yaml(filename: str):
    with open(filename) as f:
        return yaml.load(f, yaml.Loader)



# Copyright 2020-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import os.path
import typing

from .iniconfig import IniConfig

DEFAULT_CONFIG_PATH = os.path.expanduser('~/.config/dewi/config.ini')


class AppConfig(IniConfig):
    remote_mode = False

    @property
    def basedir(self) -> str:
        return self.get('core', 'basedir')

    @property
    def etcdir(self) -> str:
        return f'{self.basedir}/etc'

    @property
    def projectdir(self) -> str:
        return f'{self.basedir}/projects'

    @property
    def srcdir(self) -> str:
        return os.path.join(self.basedir, 'src')

    @property
    def tmpdir(self) -> str:
        return os.path.join(self.basedir, 'tmp')

    def repo_dir_of(self, repo_name: str) -> str:
        return os.path.join(self.srcdir, 'bare', repo_name)


_config: AppConfig = None


def get_config(path: typing.Optional[str] = None) -> AppConfig:
    global _config
    if not _config:
        config = AppConfig()
        config.open(path or DEFAULT_CONFIG_PATH)
        _config = config

    return _config


def _set_config(cfg: typing.Optional[AppConfig]):
    global _config
    _config = cfg



# Copyright 2018-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import enum
import logging
import logging.handlers
import sys
import typing

from dewi_core.config.node import Node
from dewi_core.utils.dictionaries import sort_dict


class LoggerType(enum.Enum):
    SYSLOG = enum.auto()
    CONSOLE = enum.auto()
    FILE = enum.auto()
    NONE = enum.auto()


class LogLevel(enum.Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_string(s: str):
        try:
            return LogLevel[s.upper()]
        except KeyError as exc:
            raise ValueError(exc)


class LoggerConfig(Node):
    def __init__(self):
        self.name: str = ''
        self.level: str = ''
        self.log_none: bool = False
        self.log_syslog: bool = False
        self.log_console: bool = False
        self.log_file: typing.Optional[typing.List[str]] = []

    @classmethod
    def create(cls, name: str, level: str, log_none: bool, log_syslog: bool, log_console: bool,
               log_file: typing.Optional[typing.List[str]]):
        c = cls()
        c.name = name
        c.level = level
        c.log_none = log_none
        c.log_syslog = log_syslog
        c.log_console = log_console
        c.log_file = log_file
        return c


class _Handlers:
    CONSOLE_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'
    SYSLOG_FORMAT = '%(name)s[%(process)d]: %(message)s'
    DATE_FIELD_FORMAT = '%Y-%m-%dT%H:%M:%S%z'

    @classmethod
    def create_syslog_handler(cls):
        if sys.platform == "darwin":
            address = "/var/run/syslog"
        else:
            address = ('localhost', 514)

        handler = logging.handlers.SysLogHandler(
            address=address,
            facility=logging.handlers.SysLogHandler.LOG_LOCAL0,
        )

        handler.setFormatter(logging.Formatter(cls.SYSLOG_FORMAT, datefmt=cls.DATE_FIELD_FORMAT))

        return handler

    @classmethod
    def create_console_handler(cls):
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(cls.CONSOLE_FORMAT, datefmt=cls.DATE_FIELD_FORMAT))

        return handler

    @classmethod
    def create_file_handler(cls, filename: str):
        handler = logging.FileHandler(filename)
        handler.setFormatter(logging.Formatter(cls.CONSOLE_FORMAT, datefmt=cls.DATE_FIELD_FORMAT))

        return handler

    @classmethod
    def create_null_handler(cls):
        return logging.NullHandler()

    @classmethod
    def create_handler(cls, logger_type: LoggerType, *, filename: typing.Optional[str] = None):
        if logger_type == LoggerType.CONSOLE:
            return cls.create_console_handler()
        if logger_type == LoggerType.SYSLOG:
            return cls.create_syslog_handler()
        if logger_type == LoggerType.FILE:
            return cls.create_file_handler(filename)
        # same as: logger_type == LoggerType.NONE:
        return cls.create_null_handler()


def _format_message(message: str, args: typing.Dict) -> str:
    if not args:
        return message + ';'

    return '{}; {}'.format(message, ', '.join(f'{k}={v!r}' for k, v in args.items()))


class Logger:
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR

    CRITICAL = logging.CRITICAL

    def __init__(self, name: str, logger_types: typing.List[LoggerType], *, filenames: typing.List[str] = None):
        self._logger = logging.getLogger(name)

        if LoggerType.NONE in logger_types:
            self._logger.addHandler(_Handlers.create_handler(LoggerType.NONE))
        else:
            for lt in logger_types:
                if lt == LoggerType.FILE:
                    for filename in filenames:
                        self._logger.addHandler(_Handlers.create_handler(LoggerType.FILE, filename=filename))
                else:
                    self._logger.addHandler(_Handlers.create_handler(lt))

    def set_level(self, level: LogLevel):
        self._logger.setLevel(level.value)

    def log(self, level: int, message: str, *args, **kwargs):
        if args:
            kwargs.update(args[0])
        args = sort_dict(kwargs)
        message_with_args = _format_message(message, args)
        self._logger.log(level, message_with_args)

    def debug(self, *args, **kwargs):
        self.log(self.DEBUG, *args, **kwargs)

    def info(self, *args, **kwargs):
        self.log(self.INFO, *args, **kwargs)

    def warning(self, *args, **kwargs):
        self.log(self.WARNING, *args, **kwargs)

    def error(self, *args, **kwargs):
        self.log(self.ERROR, *args, **kwargs)

    def critical(self, *args, **kwargs):
        self.log(self.CRITICAL, *args, **kwargs)

    def enabled_for(self, level: LogLevel) -> bool:
        return self._logger.isEnabledFor(level.value)


logger: Logger = None


def create_logger(name: str, logger_types: typing.Union[LoggerType, typing.List[LoggerType]], log_level: str = 'info',
                  *,
                  filenames: typing.Optional[typing.List[str]] = None):
    global logger  # pylint:disable=C0103

    if isinstance(logger_types, LoggerType):
        logger_types = [logger_types]

    logger = Logger(name, logger_types, filenames=filenames or [])
    logger.set_level(LogLevel.from_string(log_level))


def create_logger_from_config(config: LoggerConfig) -> int:
    if config.log_none:
        if config.log_syslog or config.log_file or config.log_console:
            print('ERROR: --log-none cannot be used any other log target,')
            print('ERROR: none of: --log-file, --log-console, --log-syslog')
            return 1
        create_logger(config.name, LoggerType.NONE, config.level, filenames=[])
    else:
        logger_types = []
        if config.log_console:
            logger_types.append(LoggerType.CONSOLE)
        if config.log_file:
            logger_types.append(LoggerType.FILE)
        if config.log_syslog:
            logger_types.append(LoggerType.SYSLOG)

        if not logger_types:
            # Using default logger
            logger_types = LoggerType.CONSOLE

        create_logger(config.name, logger_types, config.level, filenames=config.log_file)

    return 0


def log_debug(*args, **kwargs):
    logger.debug(*args, **kwargs)


def log_info(*args, **kwargs):
    logger.info(*args, **kwargs)


def log_warning(*args, **kwargs):
    logger.warning(*args, **kwargs)


def log_error(*args, **kwargs):
    logger.error(*args, **kwargs)


def log_critical(*args, **kwargs):
    logger.critical(*args, **kwargs)


def log_enabled_for(level: LogLevel):
    logger.enabled_for(level)


# NOTE: ensure that log_*() can be called without explicitly
# calling create_logger()
create_logger('_main_', LoggerType.NONE, 'info')



# Copyright 2015-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import sys
import traceback
import typing


def print_backtrace(tbe: typing.Optional[traceback.TracebackException] = None):
    print(format_backtrace(tbe))


def format_backtrace(tbe: typing.Optional[traceback.TracebackException] = None) -> str:
    tbe = tbe or traceback.TracebackException(*sys.exc_info())
    return f'Exception occurred:\n  Type: {tbe.exc_type.__name__}\n  Message: {tbe}\n\n' \
            + '* Details - First exception:\n' + _format_traceback_exception(tbe)


def _format_traceback_exception(exc: traceback.TracebackException, prefix: typing.Optional[str] = None):
    tb_str = ''

    if exc.__cause__ is not None:
        tb_str += _format_traceback_exception(
            exc.__cause__,
            '\n* The above exception was the direct cause of the following exception:')
    elif exc.__context__ is not None and not exc.__suppress_context__:
        tb_str += _format_traceback_exception(
            exc.__context__,
            '\n* During handling of the above exception, another exception occurred')
    return tb_str + _format_single_traceback(exc, prefix)


def _format_single_traceback(exc: traceback.TracebackException, prefix: typing.Optional[str] = None):
    tb_str = f'  Type: {exc.exc_type.__name__}\n  Message: {exc}\n\n'
    for tb in exc.stack:
        tb_str += '  File %s:%s in %s\n    %s\n' % (tb.filename, tb.lineno, tb.name, tb.line)
    if prefix:
        tb_str += f'{prefix}\n'
    return tb_str



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

from dewi_core.application import Application


def main():
    app = Application('dewi-edit-or-review')
    app.load_plugin('dewi_core.commands.edit.edit.EditPlugin')
    app.load_plugin('dewi_core.commands.review.ReviewPlugin')
    app.load_plugin('dewi_core.commands.editreview.EditReviewPlugin')
    app.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# DEWI: a developer tool and framework -- Core part
# Copyright 2012-2021  Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import sys

if sys.hexversion < 0x03080000:
    raise RuntimeError("Required python version: 3.8 or newer (current: %s)" % sys.version)

try:
    from setuptools import setup, find_packages

except ImportError:
    from distutils.core import setup

setup(
    name="dewi_core",
    description="DEWI Core: Plugin and Config Tree framework and Application",
    license="LGPLv3",
    version="5.1.1",
    author="Laszlo Attila Toth",
    author_email="python-dewi@laszloattilatoth.me",
    maintainer="Laszlo Attila Toth",
    maintainer_email="python-dewi@laszloattilatoth.me",
    keywords='tool framework development synchronization',
    url="https://github.com/LA-Toth/dewi_core",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Utilities',
    ],
    zip_safe=True,
    use_2to3=False,
    python_requires='>=3.8',
    packages=find_packages(exclude=['*tests*', 'pylintcheckers']),
    install_requires=[
        'pyyaml',
        'click >=8.0, <9',
        'click_option_group',
    ]
)



# Copyright 2016-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import collections
import typing


class _TypedDictionary(dict):
    def __init__(self, element_type: type):
        super().__init__()
        self._element_type = element_type

    def _add_element(self, key, value):
        raise NotImplementedError()

    def __setitem__(self, key, value):
        if key not in self:
            super().__setitem__(key, self._element_type())

        self._add_element(key, value)


class DictionaryWithList(_TypedDictionary):
    def __init__(self):
        super().__init__(list)

    def _add_element(self, key, value):
        self[key].append(value)


class DictionaryWithSet(_TypedDictionary):
    def __init__(self):
        super().__init__(set)

    def _add_element(self, key, value):
        self[key].add(value)


def sort_dict(d: typing.Mapping) -> collections.OrderedDict:
    result = collections.OrderedDict()
    for key in sorted(list(d.keys())):
        result[key] = d[key]

    return result



# Copyright 2015-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import importlib
import typing

from dewi_core.commandregistry import CommandRegistry
from dewi_core.config_env import ConfigDirRegistry
from dewi_core.loader.context import Context


class PluginLoaderError(Exception):
    pass


class PluginLoader:

    def __init__(self, command_registry: CommandRegistry, config_dir_registry: ConfigDirRegistry):
        self._loaded_plugins = dict()
        self._command_registry = command_registry
        self._config_dir_registry = config_dir_registry

    def load(self, plugin_names: typing.Iterable[str]) -> Context:
        dependency_graph = {}
        for name in plugin_names:
            plugin = self._get_plugin(name)
            dependency_graph[name] = plugin.get_dependencies()

        self._build_dependency_graph(dependency_graph)

        dependency_list = []
        visited_list = []
        self._build_dependency_list(dependency_graph, visited_list, dependency_list, dependency_graph.keys())

        context = Context(self._command_registry, self._config_dir_registry)
        for plugin_name in dependency_list:
            self._get_plugin(plugin_name).load(context)

        return context

    def _get_plugin(self, name: str):
        if name not in self._loaded_plugins:
            plugin = self._load_plugin(name)
            self._loaded_plugins[name] = plugin
        return self._loaded_plugins[name]

    def _load_plugin(self, name: str):
        try:
            module_name, class_name = name.rsplit('.', 1)
            module = importlib.import_module(module_name)
        except (ImportError, ValueError) as exc:
            raise PluginLoaderError(f"Plugin '{name}' is not found or cannot be imported; error='{exc}'")

        try:
            plugin_class = getattr(module, class_name)
        except AttributeError:
            raise PluginLoaderError("Plugin '{}' is not found".format(name))
        return plugin_class()

    def _build_dependency_graph(self, dependency_graph: dict):
        finished = False
        while not finished:
            changed = False
            for dependencies in dependency_graph.values():
                for dependency in dependencies:
                    if dependency not in self._loaded_plugins:
                        plugin = self._get_plugin(dependency)
                        dependency_graph[dependency] = plugin.get_dependencies()
                        changed = True
                if changed:
                    break

            if not changed:
                finished = True

    def _build_dependency_list(
            self,
            dependency_graph: dict,
            visited_nodes: list,
            dependency_list: list,
            plugin_names: typing.Iterable[str]):
        for name in plugin_names:
            if name in dependency_list:
                continue

            if name in visited_nodes:
                raise PluginLoaderError("Circular depedency in graph")
            visited_nodes.append(name)

            dependencies = dependency_graph[name]
            self._build_dependency_list(dependency_graph, visited_nodes, dependency_list, dependencies)
            dependency_list.append(name)

    @property
    def loaded_plugins(self) -> frozenset:
        return frozenset(self._loaded_plugins)



# Copyright 2017-2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import types
import typing

from dewi_core.command import Command
from dewi_core.loader.context import Context
from dewi_core.loader.plugin import Plugin


class CommandPlugin(Plugin):
    command: typing.Type[Command]

    def load(self, c: Context):
        c.commands.register_class(self.command)

    @classmethod
    def create(cls, command: typing.Type[Command]) -> typing.Type[Plugin]:
        class_name = command.__name__
        if class_name.endswith('Command'):
            class_name = class_name[:-len('Command')]
        class_name += 'Plugin'

        cls_dict = {
            'command': command
        }

        return types.new_class(class_name, (CommandPlugin,), {}, lambda ns: ns.update(cls_dict))



# Copyright 2021 Laszlo Attila Toth
# Distributed under the terms of the Apache License, Version 2.0

import os
import re
import typing

from dewi_core.appcontext import ApplicationContext
from dewi_core.command import Command
from dewi_core.optioncontext import OptionContext


def register_global_args_in_review_cmd(c: OptionContext) -> None:
    """
    Register global args not listed in the subcommands but in parent ReviewCommand.

    Added here just to be always visible in this module.
    """
    c.add_option('-U', '--gerrit-ssh-user', 'gerrit_ssh_user',
                 help='SSH username to send review into gerrit')
    c.add_option('--upload-gerrit-review', '--review', 'upload_gerrit_review', is_flag=True,
                 help='Upload gerrit review. By default the code uses dry run and print the review instead.')

    c.add_option('-R', '--repo', 'gerrit_review_local_repo', required=True,
                 help='Location of the bare git repository. '
                      'Can be an existing repo or a non-existent directory.')


class AutoReviewSubCommand(Command):
    def run(self, ctx: ApplicationContext) -> typing.Optional[int]:
        # could prepare the repository
        print("Existing repo?: ", os.path.exists(ctx.commands_args['review'].gerrit_review_local_repo))
        self.args = ctx.args
        self.ctx = ctx
        return self._run()

    def _run(self) -> typing.Optional[int]:
        pass


class GerritChangeReviewer(AutoReviewSubCommand):
    name = 'change'
    aliases = ['single-change', 'chg']
    description = 'Review a single change (commit)'

    @staticmethod
    def register_arguments(c: OptionContext):
        c.add_option('--gerrit-url', dest='gerrit_url',
                     required=True, help='Gerrit URL without revision ID')
        c.add_option('--change-id', 'change_id', type=int,
                     help='Change ID, should be the same as the end of the gerrit URL')
        c.add_option('--gerrit-revision', '--revision', 'revision', type=int,
                     required=True, help='Revision of the change in gerrit url')

    def _run(self) -> typing.Optional[int]:
        try:
            gerrit_host, change_id = self._validate_args()
        except ValueError as exc:
            print(exc)
            return 1
        print('single change', gerrit_host, change_id, self.args.change_id, self.args.revision)
        return 0

    def _validate_args(self) -> typing.Tuple[str, int]:
        if self.args.change_id and not self.args.gerrit_url.endswith(f'/{self.args.change_id}'):
            raise ValueError('Unexpected change ID, differs from the one specified in Gerrit URL')

        m = re.match(r'https://([^/]+)/[a-z/#+]+/([0-9]+)$', self.args.gerrit_url)
        if not m:
            raise ValueError('Invalid gerrit URL, should be: host/project/changeid without revision')

        return m.group(1), int(m.group(2))


class GerritChangeChainReviewer(GerritChangeReviewer):
    name = 'chain'
    aliases = []
    description = 'Review a chain of changes (commits) with the gerrit URL as HEAD'

    @staticmethod
    def register_arguments(c: OptionContext) -> None:
        GerritChangeReviewer.register_arguments(c)
        # register own arguments
        c.add_option('-r', '--repo', 'gerrit_review_local_repo', help='Conflicts with parent command, should work')

    def _run(self) -> typing.Optional[int]:
        # process related changes to the specified one, as it contains the complete chain
        gerrit_host, change_id = self._validate_args()
        print('chain', gerrit_host, change_id, self.args.change_id, self.args.revision)
        return
